/**
 * @file Controller logic for asr.xsl.
 * @author Carl Stephens
 */

import { TranscribeService, TranscriptionError, TranscriptionModel, TranscriptionMetadata } from "./TranscribeModule.js";

class TranscriptionViewModel
{
    /**
     * Initialises a new instance of the {@link TranscriptionViewModel} class.
     * 
     * @param {TranscriptionModel} transcription
     * @param {File} file The file from which the transcription was generated.
     */
    constructor(transcription, file) {
        /** @type {String} The UUID of this transcription. */
        this.id = UUID.generate();

        /** @type {String} The transcription. */
        this.transcription = transcription.transcription;

        /** @type {String} The name of the file from which the transcription was generated. */
        this.fileName = transcription.file_name;

        /** @type {TranscriptionMetadata[]} The transcription metadata. */
        this.metadata = transcription.metadata;

        /** @type {File} The file from which the transcription was generated. */
        this.file = file;

        /** @type {Boolean} Gets or sets a value indicating if the transcription view is expanded. */
        this.isExpanded = false;

        /** @type {Boolean} Gets or sets a value indicating if this transcription view has been deleted. */
        this.isDeleted = false;

        /** @type {{word: String, startTime: Number, endTime: Number, shouldHighlight: Boolean}[]} */
        this.words = [];
    }
}

class TranscriptionViewFailure
{
    /**
     * 
     * @param {String} fileName The name of the file for which this failure occured.
     * @param {String} message The reason for this failure.
     */
    constructor(fileName, message) {
        /** @type {String} The UUID of this failure. */
        this.id = UUID.generate();

        /** @type {String | null} The name of the file for which this failure occured. */
        this.fileName = fileName;

        /** @type {String} The reason for the failure. */
        this.message = message;

        /** @type {Boolean} Gets or sets a value indicating if this transcription view has been deleted. */
        this.isDeleted = false;
    }
}

/** @type {HTMLAudioElement} */
var EL_AUDIO_TRANSCRIPTION_AUDIO;

/** @type {HTMLSourceElement} */
var EL_SOURCE_TRANSCRIPTION_AUDIO;

/** @type {Number} The size of each character in our monospace font. */
var MONOSPACE_CHAR_SIZE;

/** @type {HTMLInputElement} */
var EL_INPUT_AUDIO_FILE;

var transcribeService = new TranscribeService();

window.addEventListener("load", function()
{
    // @ts-ignore | Object could be null
    EL_AUDIO_TRANSCRIPTION_AUDIO = document.getElementById("transcriptionAudio");

    // @ts-ignore | Object could be null
    EL_INPUT_AUDIO_FILE = document.getElementById("audioFileInput");

    // @ts-ignore | Object could be null
    EL_SOURCE_TRANSCRIPTION_AUDIO = document.getElementById("transcriptionAudioSource");

    const monoCharSizeTestElement = document.querySelector(".monospace-font-sizer");
    if (monoCharSizeTestElement == null || monoCharSizeTestElement.textContent == null)
    {
        MONOSPACE_CHAR_SIZE = 8; // Slightly over-estimated guess for 16px font-size.
        return;
    }
    
    MONOSPACE_CHAR_SIZE = monoCharSizeTestElement.clientWidth / monoCharSizeTestElement.textContent.length;

    pollAudioTime();
});

// Adapted from https://davidwalsh.name/javascript-polling
function pollAudioTime()
{
    var lastTime = 0;
    (function p()
    {
        if (EL_AUDIO_TRANSCRIPTION_AUDIO.currentTime != lastTime)
        {
            lastTime = EL_AUDIO_TRANSCRIPTION_AUDIO.currentTime;
            TranscriptionsListVM.currentAudioTime = lastTime;
        }
        
        setTimeout(p, 50);
    })();
}

/**
 * Fast UUID generator, RFC4122 version 4 compliant.
 * @author Jeff Ward (jcward.com).
 * @license MIT
 * @link http://stackoverflow.com/questions/105034/how-to-create-a-guid-uuid-in-javascript/21963136#21963136
 **/
 var UUID = (function()
 {
    var self = {};
    var lut = [];

    for (var i=0; i<256; i++){
        lut[i] = (i < 16 ? '0' : '') + (i).toString(16);
    }

    self.generate = function()
    {
        var d0 = Math.random() * 0xffffffff | 0;
        var d1 = Math.random() * 0xffffffff | 0;
        var d2 = Math.random() * 0xffffffff | 0;
        var d3 = Math.random() * 0xffffffff | 0;

        return lut[d0 & 0xff] + lut[d0 >> 8 & 0xff] + lut[d0 >> 16 & 0xff] + lut[d0 >> 24 & 0xff] + '-' +
            lut[d1 & 0xff] + lut[d1 >> 8 & 0xff] + '-' + lut[d1 >> 16 & 0x0f | 0x40] + lut[d1 >> 24 & 0xff] + '-' +
            lut[d2 & 0x3f | 0x80] + lut[d2 >> 8 & 0xff] + '-' + lut[d2 >> 16 & 0xff] + lut[d2 >> 24 & 0xff] +
            lut[d3 & 0xff] + lut[d3 >> 8 & 0xff] + lut[d3 >> 16 & 0xff] + lut[d3 >> 24 & 0xff];
    }

    return self;
})();
// TODO: Hash file name and size instead. Good indicator that the user has uploaded a duplicate.

// @ts-ignore
const AudioUploadComponent = Vue.createApp(
    {
        data()
        {
            return {
                files: [],
                canTranscribe: false,
                isTranscribing: false
            }
        },
        computed:
        {
            anyFiles() {
                return this.files.length > 0;
            },
            getFileNameList()
            {
                let fileNameList = "";

                for (let file of this.files) {
                    fileNameList += file.name + ", ";
                }

                return fileNameList.slice(0, fileNameList.length - 2);
            }
        },
        methods:
        {
            openFilePicker() {
                EL_INPUT_AUDIO_FILE.click(); 
            },
            onFilesChanged()
            {
                this.files = [];

                if (EL_INPUT_AUDIO_FILE.files?.length != undefined && EL_INPUT_AUDIO_FILE.files?.length > 0) {
                    
                    for (const file of EL_INPUT_AUDIO_FILE.files)
                        this.files.push(file);
                }
            },
            async doTranscription() {
                // TODO: Push files to queue. If currently transcribing, good to go; it'll pull off it.
                // Else call getTranscriptions();
                // Then, we don't have to worry about preventing the user from transcribing multiple items.
                const files = this.files;
                this.files = []; // Clear the file list, as there is no reason the user would want to transcribe the same file multiple times over
                await getTranscriptions(files);
            }
        }
    }
);
const AudioUploadVM = AudioUploadComponent.mount("#audioUploadContainer");

// @ts-ignore
const TranscriptionsListComponent = Vue.createApp(
{
    data()
    {
        return {
            /** @type {Map<String, TranscriptionViewModel>} */
            transcriptions: new Map(),
            /** @type {Map<String, TranscriptionViewFailure>} */
            failures: new Map(),
            showCharDisplay: false,
            enableTextEditing: false,
            /** @type {{id: String, url: String} | null} Gets the ID of the transcription for which the audio is currently loaded */
            currentlyLoadedAudio: null,
            currentAudioTime: 0
        }
    },
    methods:
    {
        playAudioFile(transcriptionId, startTime = 0)
        {
            if (this.enableTextEditing && startTime != -1) { // Check the start time to know if we've been triggered by a word click
                return;
            }

            // A value of -1 indicates that the play button has been used.
            // Any other causes of negative time values, for whatever unknown reason, are silently corrected
            if (startTime < 0) {
                startTime = 0;
            }

            if (startTime >= EL_AUDIO_TRANSCRIPTION_AUDIO.duration)
            {
                console.warn("Cannot start audio playback at a time longer than the audio duration.");
                return;
            }
            
            this.currentlyLoadedAudio = loadTranscriptionAudio(transcriptionId, this.currentlyLoadedAudio);

            EL_AUDIO_TRANSCRIPTION_AUDIO.currentTime = startTime;
            EL_AUDIO_TRANSCRIPTION_AUDIO.play();
        },
        removeTranscription(id)
        {
            this.transcriptions.get(id).isDeleted = true;
            delay(550).then(() => this.transcriptions.delete(id));
        },
        removeFailure(id) {
            this.failures.get(id).isDeleted = true;
            delay(550).then(() => this.failures.delete(id));
        },
        getChars(transcriptionId)
        {
            /** @type {TranscriptionModel} */
            const transcription = this.transcriptions.get(transcriptionId);
            const chars = [];

            for (const metadata of transcription.metadata)
            {
                if (metadata.char == ' ') {
                    chars.push({ char: "\u00A0", startTime: metadata.start_time });
                }
                else {
                    chars.push({ char: metadata.char, startTime: metadata.start_time });
                }
            }

            return chars;
        }
    },
    watch:
    {
        currentAudioTime(newValue)
        {
            // TODO: Replace with more efficient search. Even binary would be a significant improvement for longer transcriptions.
            const t = this.transcriptions.get(this.currentlyLoadedAudio.id);

            for (let i = 0; i < t.words.length; i++)
            {
                const word = t.words[i];

                if (word.startTime < newValue && word.endTime > newValue) {
                    word.shouldHighlight = true;
                }
                else {
                    word.shouldHighlight = false;
                }
            }
        }
    }
});
const TranscriptionsListVM = TranscriptionsListComponent.mount("#transcriptionsDisplayContainer");

/**
 * When awaited, creates a delay for the given number of milliseconds.
 * 
 * @param {Number} delayMilliseconds The number of milliseconds to delay for.
 * @returns A promise that will resolve when the delay has finished.
 */
function delay(delayMilliseconds)
{
    return new Promise(
        resolve =>
        {
            setTimeout(
                () => {
                    resolve(2);
                },
                delayMilliseconds
            );
        }
    );
}

/**
 * Gets the transcription of each submitted audio file.
 * 
 * @param {File[]} files The files to transcribe.
 */
async function getTranscriptions(files)
{
    AudioUploadVM.isTranscribing = true;
    
    // await delay(200); // TODO: Remove - UI testing purposes only
    
    // Transcribe each audio file in batches.
    try
    {
        for await (const batch of transcribeService.batchTranscribeFiles(files))
        {
            for (const t of batch)
            {
                if (!t.success) {
                    let failure = new TranscriptionViewFailure(t.file_name, t.log);
                    TranscriptionsListVM.failures.set(failure.id, failure);
                }
                else {
                    let f = files.find(f => f.name == t.file_name);
                    if (f == undefined)
                        throw new Error("File name mismatch");

                    let model = new TranscriptionViewModel(t, f);
                    model.words = getTranscriptionWords(t);

                    TranscriptionsListVM.transcriptions.set(model.id, model);
                }
            }
        }
    }
    catch (e)
    {
        console.error("Failed to transcribe files");
        console.error(e);
        let failure = new TranscriptionViewFailure(e.fileName, e.message);
        TranscriptionsListVM.failures.set(failure.id, failure);
    }
    
    AudioUploadVM.isTranscribing = false;
}

/**
 * 
 * @param {TranscriptionModel} transcription The transcription.
 * @returns {{word: String, startTime: Number, endTime: Number, shouldHighlight: Boolean}[]}
 */
function getTranscriptionWords(transcription)
{
    /** @type {{word: String, startTime: Number, endTime: Number, shouldHighlight: Boolean}[]} */
    const words = [];

    let lastWord = "";
    let currStartTime = 0;

    for (const metadata of transcription.metadata) 
    {
        if (metadata.char == ' ')
        {
            lastWord += "\u00A0";
            words.push({ word: lastWord, startTime: currStartTime, endTime: metadata.start_time, shouldHighlight: false });

            lastWord = "";
            currStartTime = metadata.start_time;
        }
        else
        {
            lastWord += metadata.char;
        }
    }

    // Push the last word, as most transcriptions will not end in a space (hence breaking the above algorithm)
    if (lastWord.length > 0) {
        words.push(
            {
                word: lastWord,
                startTime: currStartTime,
                endTime: transcription.metadata[transcription.metadata.length - 1].start_time,
                shouldHighlight: false
            });
    }

    return words;
    // console.log(this);
    // console.log(this.$refs);
    // let charsPerLine = Math.floor(this.$refs.wordListContainer.clientWidth / MONOSPACE_CHAR_SIZE);

    // return getWidthNormalisedLines(transcription.transcription, charsPerLine);
}

/**
 * Splits a transcription into line objects that fit as many words per line as possible.
 * 
 * @param {String} transcription The transcription to generate lines from.
 * @returns {{words: String[], chars: String[]}[]}
 */
function getWidthNormalisedLines(transcription, charsPerLine)
{
    /** @type {{words: String[], chars: String[]}[]} */
    let lines = [];

    if (charsPerLine < 1)
    {
        console.error("Attempted to normalise lines to zero or less characters per line.");
        return lines;
    }

    for (let i = 0; i < transcription.length; i += 0)
    {
        /** @type {String} */
        let slice;

        if (i + charsPerLine < transcription.length)
        {
            slice = transcription.slice(i, charsPerLine);

            if (transcription.charAt(i) != ' ' && transcription.charAt(i + 1) != ' ')
            {
                let lastSpacePos = slice.lastIndexOf(' ');
                let decrement = slice.length - lastSpacePos;
                slice = slice.slice(0, lastSpacePos);
                i -= decrement;
            }
        }
        else
        {
            slice = transcription.slice(i);
        }

        /** @type {{words: String[], chars: String[]}} */
        const convertedLine =
        {
            words: [],
            chars: []
        }

        for (const word of slice.split(' ')) {
            convertedLine.words.push(word + " ");
        }

        for (const char of slice) {
            convertedLine.chars.push(char);
        }

        lines.push(convertedLine);
        i += charsPerLine;
    }

    return lines;
}

/**
 * Loads an audio file.
 * @param {String} transcriptionId The name of the requested audio file.
 * @param {{id: String, url: String} | null} current The currently loaded audio.
 * @returns {{id: String, url: String}} If a new audio file was loaded, a new audio tracking object, else the current one.
 */
function loadTranscriptionAudio(transcriptionId, current)
{
    if (current == null || current.id != transcriptionId)
    {
        // TODO: Need to profile on some larger tracks; this may not be worth it? Better to cache a URL object instead?
        if (current != null) {
            URL.revokeObjectURL(current.url);
        }

        const urlObject = URL.createObjectURL(TranscriptionsListVM.transcriptions.get(transcriptionId).file);
        EL_SOURCE_TRANSCRIPTION_AUDIO.src = urlObject;
        EL_AUDIO_TRANSCRIPTION_AUDIO.load();

        return { id: transcriptionId, url: urlObject };
    }

    return current;
}
